#include "fmcReadmcpicoEvent.h"
#include "fmcEventHeader.h"
#include "fmcParticle.h"

#include <TStopwatch.h>

ClassImp(fmcReadmcpicoEvent);

fmcReadmcpicoEvent::fmcReadmcpicoEvent() : fChain(0)
{
}

fmcReadmcpicoEvent::~fmcReadmcpicoEvent()
{
    if (!fChain)
        return;
    delete fChain->GetCurrentFile();
}

Int_t fmcReadmcpicoEvent::GetEntry(Long64_t entry)
{
    // Read contents of entry.
    if (!fChain)
        return 0;
    return fChain->GetEntry(entry);
}

Long64_t fmcReadmcpicoEvent::LoadTree(Long64_t entry)
{
    // Set the environment to read one entry
    if (!fChain)
        return -5;
    Long64_t centry = fChain->LoadTree(entry);
    if (centry < 0)
        return centry;
    if (fChain->GetTreeNumber() != fCurrent)
    {
        fCurrent = fChain->GetTreeNumber();
        Notify();
    }
    return centry;
}

void fmcReadmcpicoEvent::Init(TTree *tree)
{
    // The Init() function is called when the selector needs to initialize
    // a new tree or chain. Typically here the branch addresses and branch
    // pointers of the tree will be set.
    // It is normally not necessary to make changes to the generated
    // code, but the routine can be extended by the user if needed.
    // Init() will be called many times when running on PROOF
    // (once per file to be processed).

    // Set branch addresses and branch pointers
    if (!tree)
        return;
    fChain = tree;
    fCurrent = -1;
    fChain->SetMakeClass(1);

    fChain->SetBranchAddress("bimp", &bimp, &b_bimp);
    fChain->SetBranchAddress("phi2", &phi2, &b_phi2);
    fChain->SetBranchAddress("phi3", &phi3, &b_phi3);
    fChain->SetBranchAddress("ecc2", &ecc2, &b_ecc2);
    fChain->SetBranchAddress("ecc3", &ecc3, &b_ecc3);
    fChain->SetBranchAddress("npart", &npart, &b_npart);
    fChain->SetBranchAddress("nh", &nh, &b_nh);
    fChain->SetBranchAddress("momx", momx, &b_momx);
    fChain->SetBranchAddress("momy", momy, &b_momy);
    fChain->SetBranchAddress("momz", momz, &b_momz);
    fChain->SetBranchAddress("ene", ene, &b_ene);
    fChain->SetBranchAddress("hid", hid, &b_hid);
    fChain->SetBranchAddress("pdg", pdg, &b_pdg);
    fChain->SetBranchAddress("charge", charge, &b_charge);
    Notify();
}

Bool_t fmcReadmcpicoEvent::Notify()
{
    // The Notify() function is called when a new file is opened. This
    // can be either for a new TTree in a TChain or when when a new TTree
    // is started when using PROOF. It is normally not necessary to make changes
    // to the generated code, but the routine can be extended by the
    // user if needed. The return value is currently not used.

    return kTRUE;
}

void fmcReadmcpicoEvent::Show(Long64_t entry)
{
    // Print contents of entry.
    // If entry is not specified, print current entry
    if (!fChain)
        return;
    fChain->Show(entry);
}
Int_t fmcReadmcpicoEvent::Cut(Long64_t entry)
{
    // This function may be called from Loop.
    // returns  1 if entry is accepted.
    // returns -1 otherwise.
    return 1;
}

void fmcReadmcpicoEvent::Loop()
{
    //   In a ROOT session, you can do:
    //      root> .L fmcReadmcpicoEvent.C
    //      root> fmcReadmcpicoEvent t
    //      root> t.GetEntry(12); // Fill t data members with entry number 12
    //      root> t.Show();       // Show values of entry 12
    //      root> t.Show(16);     // Read and show values of entry 16
    //      root> t.Loop();       // Loop on all entries
    //

    //     This is the loop skeleton where:
    //    jentry is the global entry number in the chain
    //    ientry is the entry number in the current Tree
    //  Note that the argument to GetEntry must be:
    //    jentry for TChain::GetEntry
    //    ientry for TTree::GetEntry and TBranch::GetEntry
    //
    //       To read only selected branches, Insert statements like:
    // METHOD1:
    //    fChain->SetBranchStatus("*",0);  // disable all branches
    //    fChain->SetBranchStatus("branchname",1);  // activate branchname
    // METHOD2: replace line
    //    fChain->GetEntry(jentry);       //read all branches
    //by  b_branchname->GetEntry(ientry); //read only this branch
    if (fChain == 0)
        return;

    Long64_t nentries = fChain->GetEntries();

    Long64_t nbytes = 0, nb = 0;
    // fmcEvent *event = nullptr;
    fmcEventHeader *header = nullptr;
    fmcInt Nparts;
    TStopwatch timer;
    timer.Start();
    for (Long64_t jentry = 0; jentry < nentries; jentry++)
    {
        Long64_t ientry = LoadTree(jentry);
        if (ientry < 0){
            std::cerr << "fmcReadermcpicoEvent::Loop: Error loading tree!" << std::endl;
            break;
        }
        nb = fChain->GetEntry(jentry);
        nbytes += nb;
        // if (Cut(ientry) < 0) continue;
        if (jentry % 1000 == 0)
            std::cout << "fmcReadermcpicoEvent::Loop: Event ["
                      << jentry << "/" << nentries << "]" << std::endl;
        
        // event = ReadEvent(jentry);
        // delete event;

        header = ReadEventHeader(jentry);
        if (header == nullptr) continue;
        Nparts = GetNparticle();
        for (int iTr=0; iTr<Nparts; iTr++)
        {
            fmcParticle *particle = ReadParticle(iTr);
            if (particle == nullptr) continue;
            if (particle->GetCharge() == 0){
                delete particle;
                continue;
            }
            delete particle;
        }
        delete header;
    }
    timer.Stop();
    timer.Print();
}

Long64_t fmcReadmcpicoEvent::GetEntries()
{
    return fChain->GetEntries();
}

Long64_t fmcReadmcpicoEvent::GetEntriesFast()
{
    return fChain->GetEntriesFast();
}

fmcInt fmcReadmcpicoEvent::GetNparticle()
{
    return nh;
}

fmcEventHeader *fmcReadmcpicoEvent::ReadEventHeader(fmcLong iev)
{
    Long64_t ientry = LoadTree(iev);
    if (ientry < 0)
        return nullptr;
    
    fmcEventHeader *header = new fmcEventHeader();
    fChain->GetEntry(iev);

    header->SetB(bimp);
    header->SetNparts(nh);
    header->SetPhi(0.);

    return header;
}

fmcParticle *fmcReadmcpicoEvent::ReadParticle(fmcInt ipart)
{
    if (ipart<0 || ipart >= nh) return nullptr;
    fmcParticle *particle = new fmcParticle();

    particle->SetPdg(pdg[ipart]);
    particle->SetCharge(charge[ipart]);
    particle->SetPxPyPzE(momx[ipart], momy[ipart], momz[ipart], ene[ipart]);

    return particle;
}

fmcEvent *fmcReadmcpicoEvent::ReadEvent(fmcLong iev)
{
    Long64_t ientry = LoadTree(iev);
    if (ientry < 0)
        return nullptr;

    fmcEvent *event = new fmcEvent();
    fmcEventHeader *header = new fmcEventHeader();
    fmcParticle *particle = nullptr;

    fChain->GetEntry(iev);

    header->SetB(bimp);
    header->SetNparts(nh);
    header->SetPhi(0.);

    event->SetHeader(header);
    delete header;

    for (int iTr = 0; iTr < nh; iTr++)
    {
        if (charge[iTr] == 0) continue;
        particle = new fmcParticle();
        particle->SetPdg(pdg[iTr]);
        particle->SetCharge(charge[iTr]);
        particle->SetPxPyPzE(momx[iTr], momy[iTr], momz[iTr], ene[iTr]);
        event->AddParticle(particle);
        delete particle;
    }

    return event;
}