#include <iostream>
#include <fstream>
#include <vector>

#include <TStopwatch.h>
#include <TString.h>
#include <TChain.h>
#include <TH2F.h>
#include <TProfile.h>
#include <TProfile2D.h>

#include "fmcTypes.h"
#include "fmcReader.h"
#include "fmcEvent.h"
#include "fmcEventHeader.h"
#include "fmcParticle.h"
#include "fmcParticleLoopInfo.h"
#include "fmcLoopProcess.h"
#include "fmcLoopOption.h"
#include "fmcQvector.h"
#include "fmcConstants.h"
#include "fmcFunctions.cxx"

#ifndef _MCINI_
const char *_TREENAME_ = "mctree";
#endif
#ifdef _MCINI_
const char *_TREENAME_ = "events";
#endif

int main(int argc, char **argv)
{
    TString inFilelistName, outFileName, refMultFileName, histMultName = "hRefMultSTAR";
    TString resFileName;
    if (argc < 9)
    {
        std::cerr << "./readFlow -i filelist -o output -refMult refMultFile -res resolutionFile" << std::endl;
        return 10;
    }
    for (int i = 1; i < argc; i++)
    {
        if (std::string(argv[i]) != "-i" &&
            std::string(argv[i]) != "-o" &&
            std::string(argv[i]) != "-refMult" &&
            std::string(argv[i]) != "-res")
        {
            std::cerr << "\nreadFlow: Unknown parameter " << i
                      << ": " << argv[i] << std::endl;
            return 11;
        }
        else
        {
            if (std::string(argv[i]) == "-i" && i != argc - 1)
            {
                inFilelistName = argv[++i];
            }
            if (std::string(argv[i]) == "-i" && i == argc - 1)
            {
                std::cerr << "\nreadFlow: Input file name was not specified!" << std::endl;
                return 20;
            }
            if (std::string(argv[i]) == "-o" && i != argc - 1)
            {
                outFileName = argv[++i];
            }
            if (std::string(argv[i]) == "-o" && i == argc - 1)
            {
                std::cerr << "\nreadFlow: Output file name was not specified!" << std::endl;
                return 21;
            }
            if (std::string(argv[i]) == "-refMult" && i != argc - 1)
            {
                refMultFileName = argv[++i];
            }
            if (std::string(argv[i]) == "-refMult" && i == argc - 1)
            {
                std::cerr << "\nreadFlow: Output file name was not specified!" << std::endl;
                return 22;
            }
            if (std::string(argv[i]) == "-res" && i != argc - 1)
            {
                resFileName = argv[++i];
            }
            if (std::string(argv[i]) == "-res" && i == argc - 1)
            {
                std::cerr << "\nreadFlow: Output file name was not specified!" << std::endl;
                return 22;
            }
        }
    }

    TChain *chain = new TChain(_TREENAME_);
    std::fstream file(inFilelistName.Data());
    std::string str;
    int nfiles = 0;

    while (std::getline(file, str))
    {
        if (chain->Add(str.c_str()))
            nfiles++;
    }

    std::cout << "Processing readFlow..." << std::endl;
    std::cout << "Found " << nfiles << " files in the filelist." << std::endl;
    fmcReader *reader = new fmcReader();
    reader->Init(chain);

    fmcLong Nevents = reader->GetEntries();
    std::cout << "readFlow: found " << Nevents << " events." << std::endl;

    // SetOptions for particle loop
    fmcLoopOption *loopOption = new fmcLoopOption();
    loopOption->ProcessRefMult();
    loopOption->ProcessQv();

    // Init particle loop info
    fmcParticleLoopInfo *loopInfo = nullptr;

    // Init particle loop process
    fmcLoopProcess *process = new fmcLoopProcess();
    process->SetLoopOptions(loopOption);
    process->SetReader(reader);

    // Prepare centrality selection
    fmc::Lim lim;
    fmcInt centrality;
    TFile *fiMult = new TFile(refMultFileName.Data(), "read");
    TH1I *hMult = (TH1I *)fiMult->Get(histMultName.Data());
    if (!hMult)
        return 30;
    fmcDouble CentEff = (histMultName == "hRefMultSTAR") ? 1. : 0.93;
    lim = fmc::CentralityBorders(hMult, CentEff);
    delete hMult;
    fiMult->Close();
    delete fiMult;

    // Prepare resolution
    std::cout << "\nReading resolution";
    TFile *fiRes = new TFile(resFileName.Data(), "read");
    TProfile *pres;
    fmcDouble res2;
    std::vector<fmcDouble> vRes[fmc::nTpcEtaGaps];
    for (int i = 0; i < fmc::nTpcEtaGaps; i++)
    {
        std::cout << ".";
        pres = (TProfile *)fiRes->Get(Form("Res2/presTPC_gap%i", i));
        if (!pres)
        {
            std::cerr << "Error: histogram " << Form("presTPC_gap%i", i)
                      << " was not found in file!" << std::endl;
            return 2;
        }
        for (int j = 1; j <= pres->GetNbinsX(); j++)
        {
            res2 = pres->GetBinContent(j);
            if (res2 >= 0)
                vRes[i].push_back(TMath::Sqrt(res2));
            else
                vRes[i].push_back(-999.);
        }
        delete pres;
    }
    std::vector<fmcDouble> vResFHCal;
    fmcDouble res, chi, chiFull, resFull;
    pres = (TProfile *)fiRes->Get(Form("Res2/presFHCal"));
    if (!pres)
    {
        std::cerr << "Error: histogram " << Form("presFHCal")
                  << " was not found in file!" << std::endl;
        return 2;
    }
    for (int j = 1; j <= pres->GetNbinsX(); j++)
    {
        res2 = pres->GetBinContent(j);
        if (res2 <= 0)
          vResFHCal.push_back(-999.);
        else
        {
            res  = TMath::Sqrt(res2);
            chi = fmc::GetChi(res, 2, 50);
            chiFull = TMath::Sqrt(2)*chi;
            resFull = fmc::GetRes(chi, 2);
            vResFHCal.push_back(resFull);
        }
    }
    delete pres;
    fiRes->Close();
    std::cout << "done!" << std::endl;

    // Manage output
    TProfile2D *pflowPtTPC[fmc::nSpecies][fmc::nTpcEtaGaps];
    TProfile2D *pflowPtFHCal[fmc::nSpecies];
    TProfile2D *pflowPtRP[fmc::nSpecies];
    for (int i = 0; i < fmc::nSpecies; i++)
    {
        for (int j = 0; j < fmc::nTpcEtaGaps; j++)
        {
            pflowPtTPC[i][j] = new TProfile2D(Form("pflowPtTPC_pid%i_gap%i", i, j),
                                              Form("pflowPtTPC_pid%i_gap%i", i, j),
                                              500, 0.1, 5.1, 10, 0, 100);
        }
        pflowPtFHCal[i] = new TProfile2D(Form("pflowPtFHCal_pid%i", i),
                                         Form("pflowPtFHCal_pid%i", i),
                                         500, 0.1, 5.1, 10, 0, 100);
        pflowPtRP[i] = new TProfile2D(Form("pflowPtRP_pid%i", i),
                                      Form("pflowPtRP_pid%i", i),
                                      500, 0.1, 5.1, 10, 0, 100);
    }

    std::cout << std::endl;

    TStopwatch timer;
    timer.Start();

    fmcInt k;
    fmcLong Nparts;
    fmcDouble phi, pt, eta, v2, PsiEP, PsiRP, PsiEPFHCal;
    fmcInt pdg, cent10;
    fmcShort charge;

    fmcQvector *qvEastTPC[fmc::nTpcEtaGaps], *qvWestTPC[fmc::nTpcEtaGaps];
    fmcQvector *qvEastRXN, *qvWestRXN;
    fmcQvector *qvEastFHCal, *qvWestFHCal;
    fmcQvector *qvFullFHCal = new fmcQvector();

    fmcEventHeader *header;
    fmcParticle *particle;

    for (fmcLong iEv = 0; iEv < Nevents; iEv++)
    {
        if (iEv % 1000 == 0)
            std::cout << "readFlow: Event [" << iEv << "/"
                      << Nevents << "]" << std::endl;

        loopInfo = process->ProcessParticleLoop(iEv);
        if (!loopInfo)
            continue;

        // Define centrality bin
        if (loopInfo->GetMultSTAR() == 0)
        {
            delete loopInfo;
            continue;
        }
        centrality = fmc::GetCentrality(lim, loopInfo->GetMultSTAR());
        cent10 = (fmcInt)(centrality / 10.);
        if ( (centrality/10.) > lim.L.size() ||
           (centrality/10.) > lim.H.size()) continue;

        // Define Qvectors
        k = 0;
        for (int i = 0; i < fmc::nTpcEtaGaps; i++)
        {
            qvEastTPC[i] = loopInfo->GetQv(k);
            k++;
            qvWestTPC[i] = loopInfo->GetQv(k);
            k++;
        }
        qvEastRXN = loopInfo->GetQv(k);
        k++;
        qvWestRXN = loopInfo->GetQv(k);
        k++;
        qvEastFHCal = loopInfo->GetQv(k);
        k++;
        qvWestFHCal = loopInfo->GetQv(k);
        qvFullFHCal->SetX(qvEastFHCal->GetX() + qvWestFHCal->GetX());
        qvFullFHCal->SetY(qvEastFHCal->GetY() + qvWestFHCal->GetY());
        qvFullFHCal->SetHarmonic(1.);

        // // Calculate flow
        header = reader->ReadEventHeader(iEv);
        Nparts = header->GetNparts();
        PsiRP = header->GetPhi();
        PsiEPFHCal = qvFullFHCal->GetPhi();// + qvWestFHCal->GetPhi();
        delete header;
        for (int iPart = 0; iPart < Nparts; iPart++)
        {
            particle = reader->ReadParticle(iPart);

            if (!particle)
                continue;
            if (particle->GetCharge() == 0)
            {
                delete particle;
                continue;
            }

            pt = particle->GetPt();
            eta = particle->GetEta();
            phi = particle->GetPhi();
            pdg = particle->GetPdg();
            charge = particle->GetCharge();

            if (pt < fmc::ptMinCut)
            {
                delete particle;
                continue;
            }
            if (TMath::Abs(eta) > fmc::TpcEtaCut)
            {
                delete particle;
                continue;
            }

            // v2(PsiRP)
            v2 = TMath::Cos(2 * (phi - PsiRP));
            pflowPtRP[0]->Fill(pt, centrality, v2);
            if (charge>0) pflowPtRP[4]->Fill(pt, centrality, v2);
            if (charge<0) pflowPtRP[8]->Fill(pt, centrality, v2);
            if (TMath::Abs(pdg) == 211)
            {
                pflowPtRP[1]->Fill(pt, centrality, v2);
                if (charge>0) pflowPtRP[5]->Fill(pt, centrality, v2);
                if (charge<0) pflowPtRP[9]->Fill(pt, centrality, v2);
            }
            if (TMath::Abs(pdg) == 321)
            {
                pflowPtRP[2]->Fill(pt, centrality, v2);
                if (charge>0) pflowPtRP[6]->Fill(pt, centrality, v2);
                if (charge<0) pflowPtRP[10]->Fill(pt, centrality, v2);
            }
            if (TMath::Abs(pdg) == 2212)
            {
                pflowPtRP[3]->Fill(pt, centrality, v2);
                if (charge>0) pflowPtRP[7]->Fill(pt, centrality, v2);
                if (charge<0) pflowPtRP[11]->Fill(pt, centrality, v2);
            }

            // v2(PsiEPFHCal)
            v2 = TMath::Cos(2 * (phi - PsiEPFHCal));
            if (vResFHCal.at(cent10) != -999.)
            {
                v2 /= vResFHCal.at((fmcInt)(centrality / 10)); 
                pflowPtFHCal[0]->Fill(pt, centrality, v2);
                if (charge>0) pflowPtFHCal[4]->Fill(pt, centrality, v2);
                if (charge<0) pflowPtFHCal[8]->Fill(pt, centrality, v2);
                if (TMath::Abs(pdg) == 211)
                {
                    pflowPtFHCal[1]->Fill(pt, centrality, v2);
                    if (charge>0) pflowPtFHCal[5]->Fill(pt, centrality, v2);
                    if (charge<0) pflowPtFHCal[9]->Fill(pt, centrality, v2);
                }
                if (TMath::Abs(pdg) == 321)
                {
                    pflowPtFHCal[2]->Fill(pt, centrality, v2);
                    if (charge>0) pflowPtFHCal[6]->Fill(pt, centrality, v2);
                    if (charge<0) pflowPtFHCal[10]->Fill(pt, centrality, v2);
                }
                if (TMath::Abs(pdg) == 2212)
                {
                    pflowPtFHCal[3]->Fill(pt, centrality, v2);
                    if (charge>0) pflowPtFHCal[7]->Fill(pt, centrality, v2);
                    if (charge<0) pflowPtFHCal[11]->Fill(pt, centrality, v2);
                }
            }

            // v2(PsiEPTPC)
            for (int i = 0; i < fmc::nTpcEtaGaps; i++)
            {
                if (eta < 0)
                    PsiEP = qvWestTPC[i]->GetPhi();
                if (eta > 0)
                    PsiEP = qvEastTPC[i]->GetPhi();
                if (TMath::Abs(eta) > fmc::TpcEtaGap.at(i))
                {
                    v2 = TMath::Cos(2 * (phi - PsiEP));
                    if (vRes[i].at(cent10) != -999.)
                    {
                        v2 /= vRes[i].at((fmcInt)(centrality / 10));
                        pflowPtTPC[0][i]->Fill(pt, centrality, v2);
                        if (charge>0) pflowPtTPC[4][i]->Fill(pt, centrality, v2);
                        if (charge<0) pflowPtTPC[8][i]->Fill(pt, centrality, v2);
                        if (TMath::Abs(pdg) == 211)
                        {
                            pflowPtTPC[1][i]->Fill(pt, centrality, v2);
                            if (charge>0) pflowPtTPC[5][i]->Fill(pt, centrality, v2);
                            if (charge<0) pflowPtTPC[9][i]->Fill(pt, centrality, v2);
                        }
                        if (TMath::Abs(pdg) == 321)
                        {
                            pflowPtTPC[2][i]->Fill(pt, centrality, v2);
                            if (charge>0) pflowPtTPC[6][i]->Fill(pt, centrality, v2);
                            if (charge<0) pflowPtTPC[10][i]->Fill(pt, centrality, v2);
                        }
                        if (TMath::Abs(pdg) == 2212)
                        {
                            pflowPtTPC[3][i]->Fill(pt, centrality, v2);
                            if (charge>0) pflowPtTPC[7][i]->Fill(pt, centrality, v2);
                            if (charge<0) pflowPtTPC[11][i]->Fill(pt, centrality, v2);
                        }
                    }
                }
            }

            delete particle;
        }

        delete loopInfo;
    }

    // Write output
    TFile *fo = new TFile(outFileName.Data(), "recreate");
    fo->mkdir("TPC_EP");
    fo->mkdir("FHCal_EP");
    fo->mkdir("RP");
    fo->cd("TPC_EP");
    for (int i = 0; i < fmc::nSpecies; i++)
    {
        for (int j = 0; j < fmc::nTpcEtaGaps; j++)
        {
            pflowPtTPC[i][j]->Write();
        }
    }
    fo->cd("FHCal_EP");
    for (int i = 0; i < fmc::nSpecies; i++)
    {
        pflowPtFHCal[i]->Write();
    }
    fo->cd("RP");
    for (int i = 0; i < fmc::nSpecies; i++)
    {
        pflowPtRP[i]->Write();
    }
    fo->Close();

    timer.Stop();
    timer.Print();
    return 0;
}
