#include <iostream>
#include <fstream>

#include <TStopwatch.h>
#include <TString.h>
#include <TChain.h>
#include <TH1I.h>

#include "fmcTypes.h"
#include "fmcReader.h"
#include "fmcEvent.h"
#include "fmcEventHeader.h"
#include "fmcParticle.h"
#include "fmcParticleLoopInfo.h"
#include "fmcLoopProcess.h"
#include "fmcLoopOption.h"
#include "fmcQvector.h"

#ifndef _MCINI_
const char *_TREENAME_ = "mctree";
#endif
#ifdef _MCINI_
const char *_TREENAME_ = "events";
#endif

int main(int argc, char **argv)
{
    TString inFilelistName, outFileName;
    if (argc < 5)
    {
        std::cerr << "./readMult -i filelist -o output" << std::endl;
        return 10;
    }
    for (int i = 1; i < argc; i++)
    {
        if (std::string(argv[i]) != "-i" &&
            std::string(argv[i]) != "-o")
        {
            std::cerr << "\nreadMult: Unknown parameter " << i
                      << ": " << argv[i] << std::endl;
            return 11;
        }
        else
        {
            if (std::string(argv[i]) == "-i" && i != argc - 1)
            {
                inFilelistName = argv[++i];
            }
            if (std::string(argv[i]) == "-i" && i == argc - 1)
            {
                std::cerr << "\nreadMult: Input file name was not specified!" << std::endl;
                return 20;
            }
            if (std::string(argv[i]) == "-o" && i != argc - 1)
            {
                outFileName = argv[++i];
            }
            if (std::string(argv[i]) == "-o" && i == argc - 1)
            {
                std::cerr << "\nreadMult: Output file name was not specified!" << std::endl;
                return 21;
            }
        }
    }

    TChain *chain = new TChain(_TREENAME_);
    std::fstream file(inFilelistName.Data());
    std::string str;
    int nfiles = 0;

    while (std::getline(file, str))
    {
        if (chain->Add(str.c_str()))
            nfiles++;
    }

    std::cout << "Processing readMult..." << std::endl;
    std::cout << "Found " << nfiles << " files in the filelist." << std::endl;
    fmcReader *reader = new fmcReader();
    reader->Init(chain);

    fmcLong Nevents = reader->GetEntries();
    if (Nevents > 500000) Nevents = 500000;
    std::cout << "readMult: found " << Nevents << " events." << std::endl;

    // SetOptions for particle loop
    fmcLoopOption *loopOption = new fmcLoopOption();
    loopOption->ProcessRefMult();

    // Init particle loop info
    fmcParticleLoopInfo *loopInfo = nullptr;

    // Init particle loop process
    fmcLoopProcess *process = new fmcLoopProcess();
    process->SetLoopOptions(loopOption);
    process->SetReader(reader);

    // Init output
    TH1I *hrefMultSTAR = new TH1I("hRefMultSTAR","refMult",2500,0.,2500.);
    TH1I *hrefMultPHENIX = new TH1I("hRefMultPHENIX","refMult",2500,0.,2500.);

    TStopwatch timer;
    timer.Start();

    for (fmcLong iEv = 0; iEv < Nevents; iEv++)
    {
        if (iEv % 1000 == 0)
            std::cout << "readMult: Event [" << iEv << "/"
                      << Nevents << "]" << std::endl;

        loopInfo = process->ProcessParticleLoop(iEv);
        if (!loopInfo) continue;
        hrefMultSTAR->Fill(loopInfo->GetMultSTAR());
        if (loopInfo->GetMultPHENIX()>0)
            hrefMultPHENIX->Fill(loopInfo->GetMultPHENIX());

        delete loopInfo;
    }

    // Write output
    TFile *fo = new TFile(outFileName.Data(),"recreate");
    fo->cd();
    hrefMultSTAR->Write();
    hrefMultPHENIX->Write();
    fo->Close();

    timer.Stop();
    timer.Print();
    return 0;
}
